from typing import List, Dict
import os
import logging
from xml.etree import ElementTree as ET
from xmlschema import XMLSchema
from database.manager import DatabaseManager
from utils.errors.import_error.xsd_scheme_not_found import XsdSchemeNotFoundError
from utils.errors.import_error.invalid_contents_error import InvalidContentsError
from utils.errors.import_error.import_error import ImportError
from models import Book, Author, BookCategory
from sqlalchemy.exc import IntegrityError

class BookImporter:
    def __init__(self):
        # Initialize the logger and schema
        self.logger = logging.getLogger(__name__)
        try:
            self.logger.debug("Opening XSD scheme in ./")
            scheme_path = os.path.join(os.path.dirname(__file__), "book_import_scheme.xsd")
            self.schema = XMLSchema(scheme_path)
        except Exception as e:
            self.logger.error("Failed to load XSD scheme")
            raise XsdSchemeNotFoundError(f"Failed to load XSD schema: {e}")

    def parse_xml(self, file_path: str) -> List[Dict[str, object]]:
        """Parses the XML file and validates it against the XSD schema."""
        try:
            tree = ET.parse(file_path)
            root = tree.getroot()

            if not self.schema.is_valid(file_path):
                raise InvalidContentsError("XML file is not valid according to XSD schema.")

            books = []
            for book_element in root.findall("book"):
                title = book_element.find("title").text
                year_published = book_element.find("year_published").text
                description = book_element.find("description").text
                isbn = book_element.find("isbn").text

                # Parse author
                author_element = book_element.find("author")
                author = {
                    "first_name": author_element.find("first_name").text,
                    "last_name": author_element.find("last_name").text
                }

                # Parse categories
                category_elements = book_element.find("categories").findall("category")
                categories = [category_element.text for category_element in category_elements]

                # Create a book dictionary with explicit types
                book = {
                    "title": title,
                    "description": description,
                    "year_published": year_published,
                    "isbn": isbn,
                    "author": author,
                    "categories": categories
                }
                books.append(book)

            return books
        except ET.ParseError as e:
            raise ImportError(f"Failed to parse XML file: {e}")

    def save_books(self, books: List[Dict[str, object]]):
        """Saves a list of books to the database."""
        try:
            with DatabaseManager.get_session() as session:
                processed_categories = {}  # Cache for processed categories by name

                for book_dict in books:
                    self.logger.debug(f"Attempting to save {book_dict['title']}")

                    # Check if the book already exists
                    existing_book = session.query(Book).filter_by(isbn=book_dict["isbn"]).first()
                    if existing_book:
                        self.logger.warning(f"ISBN {book_dict['isbn']} already exists. Skipping.")
                        continue

                    # Check or add the author
                    existing_author = session.query(Author).filter_by(
                        first_name=book_dict["author"]["first_name"],
                        last_name=book_dict["author"]["last_name"]
                    ).one_or_none()

                    if existing_author is not None:
                        self.logger.debug(f"Author {existing_author.first_name} {existing_author.last_name} already exists. Reusing.")
                        author = existing_author
                    else:
                        self.logger.debug(f"Creating new author: {book_dict['author']['first_name']} {book_dict['author']['last_name']}")
                        author = Author(
                            first_name=book_dict["author"]["first_name"],
                            last_name=book_dict["author"]["last_name"]
                        )
                        session.add(author)

                    # Handle categories
                    filtered_categories = []
                    for category_name in book_dict["categories"]:
                        if category_name in processed_categories:
                            filtered_categories.append(processed_categories[category_name])
                            continue

                        existing_category = session.query(BookCategory).filter_by(name=category_name).one_or_none()
                        if existing_category is not None:
                            self.logger.debug(f"Category {category_name} already exists. Reusing.")
                            processed_categories[category_name] = existing_category
                            filtered_categories.append(existing_category)
                        else:
                            self.logger.debug(f"Adding new category: {category_name}")
                            new_category = BookCategory(name=category_name)
                            session.add(new_category)
                            processed_categories[category_name] = new_category
                            filtered_categories.append(new_category)

                    book = Book(
                        title=book_dict["title"],
                        description=book_dict["description"],
                        year_published=book_dict["year_published"],
                        isbn=book_dict["isbn"],
                        author=author,
                        categories=filtered_categories
                    )
                    session.add(book)
                # Commit all changes
                session.commit()
        except e:
            session.rollback()
            raise ImportError(f"An error occurred when importing books: {e}") from e
        finally:
            session.close()
