import logging

from PySide6.QtWidgets import (
    QDialog, QVBoxLayout, QHBoxLayout, QLabel, QLineEdit, QTextEdit, QPushButton, QComboBox, QFormLayout, QMessageBox
)
from PySide6.QtGui import QRegularExpressionValidator
from PySide6.QtCore import QRegularExpression
from models import Book, BookStatusEnum

from database import update_book

from utils.errors.database import DatabaseError, DatabaseConnectionError, DuplicateEntryError


class BookEditor(QDialog):
    def __init__(self, book: Book = None, parent=None):
        super().__init__(parent)

        self.logger = logging.getLogger(__name__)
        self.create_layout()

        if book:
            self.logger.debug(f"Editing existing book {book.title}")
            self.book = book
            self.create_new = False
            self.fill_with_existing_data()
        else:
            self.logger.debug("Editing a new book")
            self.create_new = True

    def create_layout(self):
        self.setWindowTitle("Books")
        self.setMinimumWidth(400)

        # Create main layout
        layout = QVBoxLayout(self)

        # Form layout for book fields
        form_layout = QFormLayout()

        # Title field
        self.title_input = QLineEdit()
        form_layout.addRow("Title:", self.title_input)

        # Author field
        self.author_label = QLabel()
        form_layout.addRow("Author: ", self.author_label)

        # Description field
        self.description_input = QTextEdit()
        form_layout.addRow("Description:", self.description_input)

        # Year published field
        self.year_input = QLineEdit()
        # self.year_input.setValidator
        form_layout.addRow("Year Published:", self.year_input)

        # ISBN field
        self.isbn_input = QLineEdit()
        self.isbn_expression = QRegularExpression("\d{10}|\d{13}")
        self.isbn_validator = QRegularExpressionValidator(self.isbn_expression)
        self.isbn_input.setValidator(self.isbn_validator)
        form_layout.addRow("ISBN:", self.isbn_input)

        self.categories_input = QLineEdit()
        self.categories_input.setEnabled(False)
        form_layout.addRow("Categories: ", self.categories_input)

        layout.addLayout(form_layout)

        # Buttons
        button_layout = QHBoxLayout()

        self.save_button = QPushButton("Save")
        self.save_button.clicked.connect(self.save_book)
        button_layout.addWidget(self.save_button)

        self.cancel_button = QPushButton("Discard")
        self.cancel_button.clicked.connect(self.reject)
        button_layout.addWidget(self.cancel_button)

        layout.addLayout(button_layout)

    def fill_with_existing_data(self):
        self.title_input.setText(self.book.title)
        self.description_input.setText(self.book.description)
        self.year_input.setText(self.book.year_published)
        self.isbn_input.setText(self.book.isbn)

    def save_book(self):
        # Update book object with input values
        self.book.title = self.title_input.text()
        full_author_name = f"{self.book.author.first_name} {
            self.book.author.last_name}"
        self.author_label.setText(full_author_name)
        self.book.description = self.description_input.toPlainText()
        self.book.year_published = self.year_input.text()
        self.book.isbn = self.isbn_input.text()

        all_categories = ", ".join(
            category.name for category in self.book.categories)
        self.categories_input.setText(all_categories)

        try:
            if self.create_new:
                pass
            else:
                update_book(self.book)
            self.accept()
        except DuplicateEntryError as e:
            QMessageBox.critical(None,
                                 "ISBN is already in use",
                                 "The ISBN provided is already in use",
                                 QMessageBox.StandardButton.Ok,
                                 QMessageBox.StandardButton.NoButton)
        except DatabaseConnectionError as e:
            QMessageBox.critical(None,
                                 "Failed to save",
                                 "Could not connect to the database",
                                 QMessageBox.StandardButton.Ok,
                                 QMessageBox.StandardButton.NoButton)
        except DatabaseError as e:
            QMessageBox.critical(self.parent,
                                 "An error occured",
                                 f"Could not save the book because of the following error: {e}",
                                 QMessageBox.StandardButton.Ok,
                                 QMessageBox.StandardButton.NoButton)


__all__ = ["BookEditor"]
